/*
* light weight WS2812 lib V2.0b
*
* Controls WS2811/WS2812/WS2812B RGB-LEDs
* Author: Tim (cpldcpu@gmail.com)
*		  Tim Fischer (tim.fischer@hs-heilbronn.de)
*
* Jan 18th, 2014  v2.0b Initial Version
* Nov 29th, 2015  v2.3  Added SK6812RGBW support
* Dec 29th, 2021  v2.4	Adjustments to get code running for F_CPU between 8.5 ... 12 MHz
*						and code clean-up
*
* License: GNU GPL v2+ (see License.txt)
*/

#include "light_ws2812.h"
#include <avr/interrupt.h>
#include <avr/io.h>
#include <util/delay.h>
 
// Setleds for standard RGB 
void inline ws2812_setleds(struct cRGB *ledarray, uint16_t leds)
{
   ws2812_setleds_pin(ledarray,leds, _BV(ws2812_pin));
}

void inline ws2812_setleds_pin(struct cRGB *ledarray, uint16_t leds, uint8_t pinmask)
{
  ws2812_sendarray_mask((uint8_t*)ledarray,leds+leds+leds,pinmask);
  _delay_us(ws2812_resettime);
}

// Setleds for SK6812RGBW
void inline ws2812_setleds_rgbw(struct cRGBW *ledarray, uint16_t leds)
{
  ws2812_sendarray_mask((uint8_t*)ledarray,leds<<2,_BV(ws2812_pin));
  _delay_us(ws2812_resettime);
}

void ws2812_sendarray(uint8_t *data,uint16_t datlen)
{
  ws2812_sendarray_mask(data,datlen,_BV(ws2812_pin));
}

void inline NOPs_for_output_low()
{
    asm volatile(" "
    #if (w1_nops&1)
    w_nop1
    #endif
    #if (w1_nops&2)
    w_nop2
    #endif
    #if (w1_nops&4)
    w_nop4
    #endif
    #if (w1_nops&8)
    w_nop8
    #endif
    #if (w1_nops&16)
    w_nop16
    #endif
    );
}

void inline NOPs_for_output_high()
{
	asm volatile(" "
	#if (w2_nops&1)
	w_nop1
	#endif
	#if (w2_nops&2)
	w_nop2
	#endif
	#if (w2_nops&4)
	w_nop4
	#endif
	#if (w2_nops&8)
	w_nop8
	#endif
	#if (w2_nops&16)
	w_nop16
	#endif
	);
}

void inline NOPs_for_rest_output()
{
	asm volatile(" "
	#if (w3_nops&1)
	w_nop1
	#endif
	#if (w3_nops&2)
	w_nop2
	#endif
	#if (w3_nops&4)
	w_nop4
	#endif
	#if (w3_nops&8)
	w_nop8
	#endif
	#if (w3_nops&16)
	w_nop16
	#endif
	);
}

void inline NOPs_for_output_low_lastbit()
{
	asm volatile(" "
	#if (w4_nops&1)
	w_nop1
	#endif
	#if (w4_nops&2)
	w_nop2
	#endif
	#if (w4_nops&4)
	w_nop4
	#endif
	#if (w4_nops&8)
	w_nop8
	#endif
	#if (w4_nops&16)
	w_nop16
	#endif
	);
}

void inline write_single_color(uint8_t curbyte,uint16_t masklo,uint8_t maskhi, uint8_t ctr)
{
    asm volatile(
    // #################### Bit 0...6 ###############
    "       ldi   %0,7  \n\t"    //									## Set Variable ctr (%0) t0 7
    " 1:            \n\t"		 //									## Only label for jump back
    "       out   %2,%3 \n\t"    //  '1' [01] '0' [01] - re			## Write data bit on port 
    :	"=&d" (ctr)    :	"r" (curbyte), "I" (_SFR_IO_ADDR(ws2812_PORTREG)), "r" (maskhi), "r" (masklo)); 	NOPs_for_output_low();     asm volatile(
    "       sbrs  %1,7  \n\t"    //  '1' [03] '0' [02]				## Check whether HIGH or LOW has to be sent 
    "       out   %2,%4 \n\t"    //  '1' [--] '0' [03] - fe-low		## Clear data bit on port when LOW
    "       lsl   %1    \n\t"    //  '1' [04] '0' [04]				## Left shift of curbyte  (%1)
    :	"=&d" (ctr)    :	"r" (curbyte), "I" (_SFR_IO_ADDR(ws2812_PORTREG)), "r" (maskhi), "r" (masklo)); 	NOPs_for_output_high();     asm volatile(
    "       out   %2,%4 \n\t"    //  '1' [+1] '0' [+1] - fe-high	## Clear data bit on port 
    :	"=&d" (ctr)    :	"r" (curbyte), "I" (_SFR_IO_ADDR(ws2812_PORTREG)), "r" (maskhi), "r" (masklo)); 	NOPs_for_rest_output();     asm volatile(
    "       dec   %0    \n\t"    //  '1' [+2] '0' [+2]				## ctr (%0) = ctr -1
    "       brne  1b \n\t"       //  '1' [+3] '0' [+4]				## Jump to label '1', when ctr (%0) = 0

    // #################### Bit 7 ###############
    "       out   %2,%3 \n\t"    //  '1' [01] '0' [01] - re			## Write data bit on port 
    :	"=&d" (ctr)    :	"r" (curbyte), "I" (_SFR_IO_ADDR(ws2812_PORTREG)), "r" (maskhi), "r" (masklo)); 	NOPs_for_output_low();     asm volatile(
    "       sbrs  %1,7  \n\t"    //  '1' [03] '0' [02]				## Check whether HIGH or LOW has to be sent 
    "       out   %2,%4 \n\t"    //  '1' [--] '0' [03] - fe-low		## Clear data bit on port when LOW
    "       lsl   %1    \n\t"    //  '1' [04] '0' [04]				## Left shift of curbyte  (%1)
    :	"=&d" (ctr)     :	"r" (curbyte), "I" (_SFR_IO_ADDR(ws2812_PORTREG)), "r" (maskhi), "r" (masklo));     NOPs_for_output_high();     asm volatile(
    "       out   %2,%4 \n\t"    //  '1' [+1] '0' [+1] - fe-high	## Clear data bit on port 
    :	"=&d" (ctr)     :	"r" (curbyte), "I" (_SFR_IO_ADDR(ws2812_PORTREG)), "r" (maskhi), "r" (masklo));     NOPs_for_output_low_lastbit();	
}

void inline ws2812_sendarray_mask(uint8_t *data,uint16_t datlen,uint8_t maskhi)
{
	uint8_t curbyte,ctr,masklo;
	uint8_t sreg_prev;
  
	ws2812_DDRREG |= maskhi; // Enable output
  
	masklo	=~maskhi&ws2812_PORTREG;
	maskhi |=        ws2812_PORTREG;
  
	sreg_prev=SREG;
	cli();  
	datlen=datlen/3+1;
	
	while (datlen--) {
		curbyte=*data++;
		write_single_color(curbyte, masklo, maskhi, ctr);	// Write RED Byte on data line

		curbyte=*data++;
		write_single_color(curbyte, masklo, maskhi, ctr);	// Write GREEN Byte on data line

		curbyte=*data++;
		write_single_color(curbyte, masklo, maskhi, ctr);	// Write BLUE Byte on data line
	}
  
	SREG=sreg_prev;
}