/*
* light weight WS2812 lib V2.0b
*
* Controls WS2811/WS2812/WS2812B RGB-LEDs
* Author: 
*		Tim (cpldcpu@gmail.com)
*		Tim Fischer (tim.fischer@hs-heilbronn.de)
*
* Jan 18th, 2014  v2.0b Initial Version
* Nov 29th, 2015  v2.3  Added SK6812RGBW support
* Dec 29th, 2021  v2.4	Adjustments to get code running for F_CPU between 8.5 ... 12 MHz
*						and code clean-up
*
* License: GNU GPL v2+ (see License.txt)
*/

// ###################### DEFINEs ###############################

#ifndef LIGHT_WS2812_H_
#define LIGHT_WS2812_H_

#include <avr/io.h>
#include <avr/interrupt.h>
#include "ws2812_config.h"

#ifndef F_CPU					
	#define F_CPU 10000000UL		// Mimimum Quarz Freq: 85000000UL
#endif

#define w_zeropulse   330
#define w_onepulse    680
#define w_totalperiod 1200

/*
 * Internal defines
 */

#define CONCAT(a, b)            a ## b
#define CONCAT_EXP(a, b)   CONCAT(a, b)

#define ws2812_PORTREG  CONCAT_EXP(PORT,ws2812_port)
#define ws2812_DDRREG   CONCAT_EXP(DDR,ws2812_port)

#endif	/* LIGHT_WS2812_H_ */
		// The only critical timing parameter is the minimum pulse length of the "0"
		// Warn or throw error if this timing can not be met with current F_CPU settings.
#define w_lowtime ((w1_nops+w_fixedlow)*1000000)/(F_CPU/1000)
#if w_lowtime>550
	#error "Light_ws2812: Sorry, the clock speed is too low. Did you set F_CPU correctly?"
#elif w_lowtime>450
	#warning "Light_ws2812: The timing is critical and may only work on WS2812B, not on WS2812(S)."
	#warning "Please consider a higher clockspeed, if possible"
#endif   

#define w_nop1  "nop      \n\t"
#define w_nop2  "rjmp .+0 \n\t"
#define w_nop4  w_nop2 w_nop2
#define w_nop8  w_nop4 w_nop4
#define w_nop16 w_nop8 w_nop8

// Timing in ns
// Fixed cycles used by the inner loop
#define w_fixedlow    2
#define w_fixedhigh   4
#define w_fixedtotal  8   

// Insert NOPs to match the timing, if possible
#define w_zerocycles    (((F_CPU/1000)*w_zeropulse          )/1000000)
#define w_onecycles     (((F_CPU/1000)*w_onepulse    +500000)/1000000)
#define w_totalcycles   (((F_CPU/1000)*w_totalperiod +500000)/1000000)

// Calculate timing between edges
#define w1 (w_zerocycles-w_fixedlow)			// w1   nops between rising edge and falling edge - low
#define w2 (w_onecycles-w_fixedhigh-w1)			// w2   nops between fe low and fe high
#define w3 (w_totalcycles-w_fixedtotal-w1-w2)	// w3   nops to complete loop

#if w1>0
	#define w1_nops w1
#else
	#define w1_nops  0
#endif

#if w2>0
	#define w2_nops w2
#else
	#define w2_nops  0
#endif

#define w3_nops w3

#if w3>4
	#define w4_nops w3-4
#else
	#define w4_nops  0
#endif

// ###################### STRUCTs ###############################

/*
 *  Structure of the LED array
 *
 * cRGB:     RGB  for WS2812S/B/C/D, SK6812, SK6812Mini, SK6812WWA, APA104, APA106
 * cRGBW:    RGBW for SK6812RGBW
 */ 

struct cRGB  { uint8_t g; uint8_t r; uint8_t b; };
struct cRGBW { uint8_t g; uint8_t r; uint8_t b; uint8_t w;};

// ###################### DECLARATIONs ###############################

/* User Interface
 * 
 * Input:
 *         ledarray:           An array of GRB data describing the LED colors
 *         number_of_leds:     The number of LEDs to write
 *         pinmask (optional): Bitmask describing the output bin. e.g. _BV(PB0)
 *
 * The functions will perform the following actions:
 *         - Set the data-out pin as output
 *         - Send out the LED data 
 *         - Wait 50s to reset the LEDs
 */

void ws2812_setleds     (struct cRGB  *ledarray, uint16_t number_of_leds);
void ws2812_setleds_pin (struct cRGB  *ledarray, uint16_t number_of_leds,uint8_t pinmask);
void ws2812_setleds_rgbw(struct cRGBW *ledarray, uint16_t number_of_leds);

/* 
 * Internal functions
 *
 * The functions take a byte-array and send to the data output as WS2812 bitstream.
 * The length is the number of bytes to send - three per LED.
 */

void ws2812_sendarray     (uint8_t *array,uint16_t length);
void ws2812_sendarray_mask(uint8_t *array,uint16_t length, uint8_t pinmask);